import os
import sys

sys.path.extend([os.path.dirname(os.getcwd()), os.path.join(os.path.dirname(os.getcwd()), "jaxparrow")])
autodoc_member_order = "groupwise"
exclude_patterns = ["_build", "**tests**"]

from version import __version__


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "jaxparrow"
copyright = "2023, Vadim Bertrand, Victor E V Z De Almeida, Julien Le Sommer, Emmanuel Cosme"
author = "Vadim Bertrand, Victor E V Z De Almeida, Julien Le Sommer, Emmanuel Cosme"
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "myst_parser"
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_js_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.1/MathJax.js?config=TeX-AMS-MML_HTMLorMML"
]
rst_prolog = """
:github_url: https://github.com/meom-group/jaxparrow
"""



import jax.numpy as jnp
from jaxtyping import Array, Float

from .geometry import compute_spatial_step, compute_coriolis_factor
from .operators import derivative, interpolation
from .sanitize import init_land_mask, sanitize_data


def advection(
        u: Float[Array, "lat lon"],
        v: Float[Array, "lat lon"],
        dx_u: Float[Array, "lat lon"],
        dy_u: Float[Array, "lat lon"],
        dx_v: Float[Array, "lat lon"],
        dy_v: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"]
) -> [Float[Array, "lat lon"], Float[Array, "lat lon"]]:
    """
    Computes the advection terms of a 2d velocity field, on a C-grid, following NEMO convention [1]_.

    Parameters
    ----------
    u : Float[Array, "lat lon"]
        U component of the velocity field (on the U grid)
    v : Float[Array, "lat lon"]
        V component of the SSC velocity field (on the V grid)
    dx_u : Float[Array, "lat lon"]
        Spatial steps on the U grid along `x`, in meters
    dy_u : Float[Array, "lat lon"]
        Spatial steps on the U grid along `y`, in meters
    dx_v : Float[Array, "lat lon"]
        Spatial steps on the V grid along `x`, in meters
    dy_v : Float[Array, "lat lon"]
        Spatial steps on the V grid along `y`, in meters
    mask : Float[Array, "lat lon"]
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)

    Returns
    -------
    u_adv_v : Float[Array, "lat lon"]
        U component of the advection term, on the V grid
    v_adv_u : Float[Array, "lat lon"]
        V component of the advection term, on the U grid
    """
    u_adv_v = _u_advection_v(u, v, dx_v, dy_v, mask)
    v_adv_u = _v_advection_u(u, v, dx_u, dy_u, mask)

    return u_adv_v, v_adv_u


def _u_advection_v(
        u_u: Float[Array, "lat lon"],
        v_v: Float[Array, "lat lon"],
        dx_u: Float[Array, "lat lon"],
        dy_u: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"]
) -> Float[Array, "lat lon"]:
    dudx_t = derivative(u_u, dx_u, axis=1, padding="left")   # (U(i), U(i+1)) -> T(i+1)
    dudx_v = interpolation(dudx_t, axis=0, padding="right")  # (T(j), T(j+1)) -> V(j)

    dudy_f = derivative(u_u, dy_u, axis=0, padding="right")  # (U(j), U(j+1)) -> F(j)
    dudy_v = interpolation(dudy_f, axis=1, padding="left")   # (F(i), F(i+1)) -> V(i+1)

    u_t = interpolation(u_u, axis=1, padding="left")   # (U(i), U(i+1)) -> T(i+1)
    u_v = interpolation(u_t, axis=0, padding="right")  # (T(j), T(j+1)) -> V(j)

    u_adv_v = u_v * dudx_v + v_v * dudy_v  # V(j)
    u_adv_v = sanitize_data(u_adv_v, 0., mask)

    return u_adv_v


def _v_advection_u(
        u_u: Float[Array, "lat lon"],
        v_v: Float[Array, "lat lon"],
        dx_v: Float[Array, "lat lon"],
        dy_v: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"]
) -> Float[Array, "lat lon"]:
    dvdx_f = derivative(v_v, dx_v, axis=1, padding="right")  # (V(i), V(i+1)) -> F(i)
    dvdx_u = interpolation(dvdx_f, axis=0, padding="left")   # (F(j), F(j+1)) -> U(j+1)

    dvdy_t = derivative(v_v, dy_v, axis=0, padding="left")   # (V(j), V(j+1)) -> T(j+1)
    dvdy_u = interpolation(dvdy_t, axis=1, padding="right")  # (T(i), T(i+1)) -> U(i)

    v_t = interpolation(v_v, axis=0, padding="left")   # (V(j), V(j+1)) -> T(j+1)
    v_u = interpolation(v_t, axis=1, padding="right")  # (T(i), T(i+1)) -> U(i)

    v_adv_u = u_u * dvdx_u + v_u * dvdy_u  # U(i)
    v_adv_u = sanitize_data(v_adv_u, 0., mask)

    return v_adv_u


def magnitude(
        u: Float[Array, "lat lon"],
        v: Float[Array, "lat lon"],
        interpolate: bool = True
) -> Float[Array, "lat lon"]:
    """
    Computes the magnitude (azimuthal velocity) of a 2d velocity field,
    possibly on a C-grid (following NEMO convention [1]_) if ``interpolate=True``.

    Parameters
    ----------
    u : Float[Array, "lat lon"]
        U component of the velocity field (on the U or T grid)
    v : Float[Array, "lat lon"]
        V component of the velocity field (on the V or T grid)
    interpolate : bool, optional
        If `True`, the velocity components are assumed to be located on the U and V grids,
        and are interpolated to the T one (following NEMO convention [1]_).
        If `False`, the velocity components are assumed to be located on the T grid, and interpolation is not needed.

        Defaults to `True`

    Returns
    -------
    magn_t : Float[Array, "lat lon"]
        Magnitude of the velocity field, on the T grid
    """
    if interpolate:
        # interpolate to the T point
        u_t = interpolation(u, axis=1, padding="left")  # (U(i), U(i+1)) -> T(i+1)
        v_t = interpolation(v, axis=0, padding="left")  # (V(j), V(j+1)) -> T(j+1)
    else:
        u_t, v_t = u, v

    magn_t = jnp.sqrt(u_t ** 2 + v_t ** 2)

    return magn_t


def normalized_relative_vorticity(
        u: Float[Array, "lat lon"],
        v: Float[Array, "lat lon"],
        lat_u: Float[Array, "lat lon"],
        lon_u: Float[Array, "lat lon"],
        lat_v: Float[Array, "lat lon"],
        lon_v: Float[Array, "lat lon"],
        mask: Float[Array, "lat lon"] = None,
        interpolate: bool = True
) -> Float[Array, "lat lon"]:
    """
    Computes the normalised relative vorticity of a velocity field, on a C-grid, following NEMO convention [1]_.

    The ``lat_u``, ``lon_u``, ``lat_v``, and ``lon_v`` are expected to follow the NEMO convention [1]_.
    If not, the function will return inaccurate results.

    Parameters
    ----------
    u : Float[Array, "lat lon"]
        U component of the velocity field (on the U grid)
    v : Float[Array, "lat lon"]
        V component of the velocity field (on the V grid)
    lat_u : Float[Array, "lat lon"]
        Latitudes of the U grid
    lon_u : Float[Array, "lat lon"]
        Longitudes of the U grid
    lat_v : Float[Array, "lat lon"]
        Latitudes of the V grid
    lon_v : Float[Array, "lat lon"]
        Longitudes of the V grid
    mask : Float[Array, "lat lon"], optional
        Mask defining the marine area of the spatial domain; `1` or `True` stands for masked (i.e. land)
    interpolate : bool, optional
        If `True`, the relative normalized vorticity is interpolated from the F grid to the T grid.
        If `False`, it remains on the F grid.

        Defaults to `True`

    Returns
    -------
    w : Float[Array, "lat lon"]
        The normalised relative vorticity,
        on the F grid (if ``interpolate=False``), or the T grid (if ``interpolate=True``)
    """
    # Make sure the mask is initialized
    mask = init_land_mask(u, mask)

    # Compute spatial step and Coriolis factor
    _, dy_u = compute_spatial_step(lat_u, lon_u)
    dx_v, _ = compute_spatial_step(lat_v, lon_v)
    f_u = compute_coriolis_factor(lat_u)

    # Handle spurious data and apply mask
    dy_u = sanitize_data(dy_u, jnp.nan, mask)
    dx_v = sanitize_data(dx_v, jnp.nan, mask)
    f_u = sanitize_data(f_u, jnp.nan, mask)

    # Compute the normalized relative vorticity
    du_dy_f = derivative(u, dy_u, axis=0, padding="right")  # (U(j), U(j+1)) -> F(j)
    dv_dx_f = derivative(v, dx_v, axis=1, padding="right")  # (V(i), V(i+1)) -> F(i)
    f_f = interpolation(f_u, axis=0, padding="right")  # (U(j), U(j+1)) -> F(j)
    w_f = (dv_dx_f - du_dy_f) / f_f  # F(j)

    if interpolate:
        w_u = interpolation(w_f, axis=0, padding="left")  # (F(j), F(j+1)) -> U(j+1)
        w = interpolation(w_u, axis=1, padding="left")  # (U(i), U(i+1)) -> T(i+1)
    else:
        w = w_f

    w = sanitize_data(w, jnp.nan, mask)

    return w


def kinetic_energy(
        u: Float[Array, "lat lon"],
        v: Float[Array, "lat lon"],
        interpolate: bool = True
) -> Float[Array, "lat lon"]:
    """
    Computes the Kinetic Energy (KE) of a velocity field,
    possibly on a C-grid (following NEMO convention [1]_) if ``interpolate=True``.

    Parameters
    ----------
    u : Float[Array, "lat lon"]
        U component of the velocity field (on the U grid)
    v : Float[Array, "lat lon"]
        V component of the velocity field (on the V grid)
    interpolate : bool, optional
        If `True`, the velocity components are assumed to be located on the U and V grids,
        and are interpolated to the T one (following NEMO convention [1]_).
        If `False`, the velocity components are assumed to be located on the T grid, and interpolation is not needed.

        Defaults to `True`

    Returns
    -------
    eke : Float[Array, "lat lon"]
        The Kinetic Energy on the T grid
    """
    if interpolate:
        # interpolate to the T point
        u_t = interpolation(u, axis=1, padding="left")  # (U(i), U(i+1)) -> T(i+1)
        v_t = interpolation(v, axis=0, padding="left")  # (V(j), V(j+1)) -> T(j+1)
    else:
        u_t, v_t = u, v

    eke_t = (u_t ** 2 + v_t ** 2) / 2

    return eke_t



from functools import partial

from cartopy import crs as ccrs
from matplotlib.animation import FuncAnimation
import matplotlib.pyplot as plt
import numpy as np


def plot_ssh(
        ds,
        ti=0,  # time index
        longitude_name="longitude",  # variables names
        latitude_name="latitude",
        ssh_name="adt",
        label=None,
        ax=None
):
    if ax is None:
        _, ax = plt.subplots(subplot_kw={"projection": ccrs.PlateCarree()})
        
    im = ax.pcolormesh(
        ds[longitude_name], ds[latitude_name], ds[ssh_name][ti],
        cmap="turbo", shading="auto",
        vmin=ds[ssh_name].min(), vmax=ds[ssh_name].max(),
        transform=ccrs.PlateCarree()
    )
    
    clb = plt.colorbar(im, ax=ax)
    clb.ax.set_title("SSH (m)")
    
    ax.coastlines()
    if label is not None:
        ax.set_title(label)
        
    return im


def plot_currents(
        ds,
        ti=0,  # time index
        longitude_name="longitude",  # variables names
        latitude_name="latitude",
        magnitude_name="uv",
        vmin=None,
        vmax=None,
        label=None,
        ax=None
):
    if ax is None:
        _, ax = plt.subplots(subplot_kw={"projection": ccrs.PlateCarree()})
        
    if vmin is None:
        vmin=ds[magnitude_name].min()
    if vmax is None:
        vmax=ds[magnitude_name].max()
        
    im = ax.pcolormesh(
        ds[longitude_name], ds[latitude_name], ds[magnitude_name][ti],
        cmap="viridis", shading="auto",
        vmin=vmin, vmax=vmax,
        transform=ccrs.PlateCarree()
    )

    clb = plt.colorbar(im, ax=ax)
    clb.ax.set_title("$\\vert\\vert \\vec{u} \\vert\\vert$ (m/s)")
    
    ax.coastlines()
    if label is not None:
        ax.set_title(label)
        
    return im


def update_ssh(im, ds, ti, ssh_name="adt"):
    im.set_array(ds[ssh_name][ti])
    return im


def update_currents(im, ds, ti, magnitude_name="uvgos"):
    im.set_array(ds[magnitude_name][ti])
    return im


class Animated:
    def __init__(
            self,
            ds,
            plot_vars,  # iterable, control what to plot
            init_fns,
            update_fns,
            ti0=None,  # starting time index
            ds_dt=np.timedelta64(1, "D"),  # temporal sampling (np.timedelta or similar)
            frame_dt=50,  # time interval between animation frames (in milliseconds)
            longitude_name="longitude",  # variables names
            latitude_name="latitude",
            time_name="time"
    ):
        if ti0 is None:
            ti0 = 0

        self.ds = ds
        self.plot_vars = plot_vars
        self.update_fns = update_fns

        time_da = ds[time_name]
        res_dt = time_da[1].data - time_da[0].data
        self.time = np.arange(time_da.min().data, time_da.max().data + res_dt, res_dt)
        t_step = 1
        if ds_dt > res_dt:  # resample if needed
            t_step = ds_dt // res_dt
        ti = np.arange(0, self.time.size, t_step)

        self.fig, ax = plt.subplots(1, len(self.plot_vars), figsize=(15, 4),
                                    subplot_kw={"projection": ccrs.PlateCarree()})

        self.artists = [
            init_fns[i](self.ds, ti0, longitude_name, latitude_name, self.plot_vars[i], ax=ax[i])
            for i in range(len(self.plot_vars))
        ]

        self.fig.suptitle(self.time[0].astype("datetime64[s]").item().strftime("%Y-%m-%d %H:%M"))
        self.fig.tight_layout()

        self.animation = FuncAnimation(
            fig=self.fig, func=self.update, blit=True,
            frames=ti, interval=frame_dt,
            cache_frame_data=False
        )

    def update(self, ti):
        artists = [
            self.update_fns[i](self.artists[i], self.ds, ti, self.plot_vars[i])
            for i in range(len(self.plot_vars))
        ]
        self.fig.suptitle(self.time[ti].astype("datetime64[s]").item().strftime("%Y-%m-%d %H:%M"))

        artists.append(self.fig)
        return artists


class AnimatedSSHCurrent(Animated):
    def __init__(
            self,
            ds,
            plot_vars,
            labels=(None, None),
            ti0=None,
            ds_dt=np.timedelta64(1, "D"),
            frame_dt=50,
            longitude_name="longitude",
            latitude_name="latitude",
            time_name="time"
    ):
        init_fns = [
            partial(plot_ssh, label=labels[0]),
            partial(plot_currents, label=labels[1]),
        ]
        update_fns = [update_ssh, update_currents]
        
        super().__init__(
            ds,
            plot_vars,
            init_fns,
            update_fns,
            ti0,
            ds_dt,
            frame_dt,
            longitude_name,
            latitude_name,
            time_name
        )


class AnimatedCurrents(Animated):
    def __init__(
            self,
            ds,
            plot_vars,
            labels,
            ti0=None,
            ds_dt=np.timedelta64(1, "D"),
            frame_dt=50,
            longitude_name="longitude",
            latitude_name="latitude",
            time_name="time"
    ):
        vmin = np.min([
            ds[plot_var].min() for plot_var in plot_vars
        ])
        vmax = np.max([
            ds[plot_var].max() for plot_var in plot_vars
        ])

        init_fns = [
            partial(plot_currents, vmin=vmin, vmax=vmax, label=labels[i]) for i in range(len(labels))
        ]
        update_fns = [update_currents] * len(labels)
        
        super().__init__(
            ds,
            plot_vars,
            init_fns,
            update_fns,
            ti0,
            ds_dt,
            frame_dt,
            longitude_name,
            latitude_name,
            time_name
        )



import argparse
from datetime import datetime
import yaml

import numpy as np
import numpy.ma as ma
import xarray as xr

from .version import __version__
from .cyclogeostrophy import cyclogeostrophy


def _read_data(
        conf_path: str
) -> list:
    with open(conf_path) as f:
        conf = yaml.safe_load(f)  # parse conf file

    values = []
    seen_ds = {}

    # variable descriptions to be found in the conf file
    variables = ["ssh", "lon", "lat", "mask"]
    for var in variables:
        try:
            conf_entry = conf[var]

            # each variable refers to a netCDF file (xarray Dataset)
            if conf_entry["file_path"] not in seen_ds:
                seen_ds[conf_entry["file_path"]] = xr.open_dataset(conf_entry["file_path"])
            ds = seen_ds[conf_entry["file_path"]]
            # and to a xarray Dataset variable
            val = ds[conf_entry["var_name"]]
            # optionally, the user can use indexing (if one needs to extract observation at a specific time for ex.)
            if "index" in conf_entry:
                idx = conf_entry["index"]
                val = val[{val.dims[i]: idx[i] for i in range(len(idx))}]
        except KeyError as e:
            if "mask" in var:  # mask variables are optional
                val = None
            else:
                raise e

        values.append(val)

    # in addition, the user can provide arguments passed to the cyclogeostrophic method
    values.append(conf.get("cyclogeostrophy", {}))
    # and, optionally tune the output netCFD metadata attributes
    values.append(conf.get("out_attrs", {}))
    # finally, he must provide the full path (including its name and extension) of the output file
    values.append(conf["out_path"])

    return values


def _reverse_masks(
        mask_ssh: np.ndarray,
        mask_u: np.ndarray,
        mask_v: np.ndarray
) -> [np.ndarray, np.ndarray, np.ndarray]:
    def do_reverse(mask):
        if mask is not None:
            return 1 - mask
    return do_reverse(mask_ssh), do_reverse(mask_u), do_reverse(mask_v)


def _apply_masks(
        u_geos: np.ndarray,
        v_geos: np.ndarray,
        u_cyclo: np.ndarray,
        v_cyclo: np.ndarray,
        mask: np.ndarray
) -> [np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    def do_apply_mask(arr):
        if mask is not None:
            return ma.masked_array(arr, mask)
        else:
            return arr
    return (do_apply_mask(u_geos), do_apply_mask(v_geos),
            do_apply_mask(u_cyclo), do_apply_mask(v_cyclo))


def _create_attrs(
        conf_path: str,
        out_attrs: dict,
        run_datetime: str
) -> dict:
    with open(conf_path) as f:
        raw_conf = f.read()

    attrs = {
        "Conventions": "CF-1.10",
        "title": "ocean geostrophic and cyclogeostrophic currents",
        "institution": "",
        "source": "jaxparrow package - v" + __version__,
        "history": run_datetime + ": jaxparrow --conf_file " + conf_path,
        "references": "https://jaxparrow.readthedocs.io/",
        "comment": "jaxparrow computes the inversion of the cyclogeostrophic balance based on a variational formulation"
                   " approach, using JAX. ",
        "conf_content": raw_conf
    }
    attrs.update(out_attrs)

    return attrs


def _to_dataset(
        u_geos: ma.MaskedArray,
        v_geos: ma.MaskedArray,
        u_cyclo: ma.MaskedArray,
        v_cyclo: ma.MaskedArray,
        lat_u: ma.MaskedArray,
        lon_u: ma.MaskedArray,
        lat_v: ma.MaskedArray,
        lon_v: ma.MaskedArray,
        conf_path: str,
        out_attrs: dict,
        run_datetime: str
) -> xr.Dataset:
    ds = xr.Dataset({
        "u_geos": (["y", "x"], u_geos),
        "v_geos": (["y", "x"], v_geos),
        "u_cyclo": (["y", "x"], u_cyclo),
        "v_cyclo": (["y", "x"], v_cyclo)
    }, coords={
        "u_lon": (["y", "x"], lon_u),
        "u_lat": (["y", "x"], lat_u),
        "v_lon": (["y", "x"], lon_v),
        "v_lat": (["y", "x"], lat_v),
    }, attrs=_create_attrs(conf_path, out_attrs, run_datetime))
    return ds


def _write_data(
        u_geos: ma.MaskedArray,
        v_geos: ma.MaskedArray,
        u_cyclo: ma.MaskedArray,
        v_cyclo: ma.MaskedArray,
        lat_u: ma.MaskedArray,
        lon_u: ma.MaskedArray,
        lat_v: ma.MaskedArray,
        lon_v: ma.MaskedArray,
        conf_path: str,
        out_attrs: dict,
        run_datetime: str,
        out_path: str
):
    ds = _to_dataset(u_geos, v_geos, u_cyclo, v_cyclo, lat_u, lon_u, lat_v, lon_v, conf_path, out_attrs, run_datetime)
    ds.to_netcdf(out_path)


def _main(
        conf_path: str
):
    run_datetime = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    ssh_t, lon_t, lat_t, mask, cyclo_kwargs, out_attrs, out_path = _read_data(conf_path)

    mask = 1 - mask  # NEMO convention: 1 means marine area ; for ma.masked_array: 1 means masked (i.e. land)

    u_cyclo_u, v_cyclo_v, u_geos_u, v_geos_v, lat_u, lon_u, lat_v, lon_v = (
        cyclogeostrophy(ssh_t, lat_t, lon_t, mask, return_geos=True, **cyclo_kwargs))

    u_cyclo_u, v_cyclo_v, u_geos_u, v_geos_v = _apply_masks(u_cyclo_u, v_cyclo_v, u_geos_u, v_geos_v, mask)

    _write_data(u_cyclo_u, v_cyclo_v, u_geos_u, v_geos_v, lat_u, lon_u, lat_v, lon_v, conf_path, out_attrs,
                run_datetime, out_path)


def main():
    parser = argparse.ArgumentParser(prog="Cyclogeostrophic balance",
                                     description="Computes the inversion of the cyclogeostrophic balance using a "
                                                 "variational or iterative approach.",
                                     epilog="For an example yaml configuration file, see the documentation: "
                                            "https://meom-group.github.io/jaxparrow/description.html#as-an-executable")

    parser.add_argument("--conf_path", type=str, help="yaml configuration file path", required=True)

    args = parser.parse_args()

    _main(args.conf_path)


if __name__ == "__main__":
    main()
